#include <vector>
#include <string>
#include <iostream>
#include <algorithm>
#include <stdlib.h>

using std::vector;
using std::string;
using std::ostream;
using std::cerr;
using std::endl;
using std::find;


template <class T> inline unsigned int index_of(const vector<T> &v, T k) {
	unsigned int index = find(v.begin(), v.end(), k) - v.begin();
	if (index==v.size())
		return -1;
	return index;
}



class ACOptions 
{
public:
	ACOptions();
	int parse_options (int argc, char * argv[]);
	void print_usage(ostream & outstream);
	
	void add_option_strings(int ID, const char *short_str, const char *long_str, const char *help);

private:
	int get_option_code(string opt);
	
public:
	string version;
	
	enum OPTIONS {OPT_HELP=0, OPT_FILENAME, OPT_SCORETYPE, OPT_SEARCHSTRATEGY, OPT_K, OPT_STOP, \
		OPT_BEAMWIDTH, OPT_ITERATIONS, OPT_LAMBDA, OPT_CYCLES, OPT_PARTITION, OPT_OUTPUT, OPT_SCORES, OPT_VERBOSE, OPT_QUIET, OPT_CNT};
	
	vector<string> short_opts, long_opts, help_opts;

	string  filename;
	int     k;
	bool    stop;
	int     scoretype;
	int     searchstrategy;
	int     beamwidth;
	int     iterations;
	double  lambda;
	int     cycles;
	string  output;
	string  graph;
	string  scores;
	string  partitionfilename;
	int    verbosity;
	bool    quiet;
	
};

ACOptions::ACOptions()
{
	short_opts = vector<string>(OPT_CNT);
	long_opts = vector<string>(OPT_CNT);
	help_opts = vector<string>(OPT_CNT);
	
	add_option_strings(OPT_FILENAME,       "-f", "--file", "data filename");
	add_option_strings(OPT_SCORETYPE,      "-t", "--type", "type of score [0=entropy, 1=twopart MDL, 2=semi-refined MDL, 3=refined MDL, 4=BIC, 5=AIC; default=1]");
	add_option_strings(OPT_SEARCHSTRATEGY, "-s", "--search", "search strategy [1=hierarchical, 2=anneal, 3=beam search; default=1]");
	add_option_strings(OPT_K,              "-k", "--k", "number of clusters to stop at [default=1]");
	add_option_strings(OPT_STOP,           "-x", "--stop", "stop when in a local minimum");
	add_option_strings(OPT_BEAMWIDTH,      "-b", "--beamwidth", "beam width [default=5]");
	add_option_strings(OPT_ITERATIONS,     "-i", "--iterations", "number of steps in simulated annealing [default=1000]");
	add_option_strings(OPT_LAMBDA,         "-l", "--lambda", "simulated annealing scaling factor [default=1]");
	add_option_strings(OPT_CYCLES,         "-c", "--cycles", "simulated annealing nr of cycles [default=10]");
	add_option_strings(OPT_PARTITION,      "-p", "--partition", "compute score of partition in file");
	add_option_strings(OPT_OUTPUT,         "-o", "--output", "final partition output filename");
	add_option_strings(OPT_SCORES,         "-a", "--scores", "store all intermediate scores in a file");
	add_option_strings(OPT_VERBOSE,        "-v", "--verbose", "verbosity level [default=1]");
	add_option_strings(OPT_QUIET,          "-q", "--quiet", "quiet mode [-v 0]");
	
	
	scoretype      = 1; // AttrClusterer::TWOPART;
	searchstrategy = 1; // AttrClusterer::HIERARCHICAL;
	k			   = 1;
	stop		   = false;
	beamwidth      = 5;
	iterations     = 1000;
	lambda         = 1.0;
	cycles         = 10;
	verbosity      = 1;
	
	return;
}

void ACOptions::add_option_strings(int ID, const char *short_str, const char *long_str, const char *help_str)
{
	short_opts[ID] = short_str;
	long_opts[ID]  = long_str;
	help_opts[ID]  = help_str;
	
	return;
}	

int ACOptions::get_option_code(string opt)
{
	int option = index_of(short_opts, opt);
	if (option == -1)
		option = index_of(long_opts, opt);
	
	return option;
}


void ACOptions::print_usage(ostream &outstream)
{
	outstream << "USAGE:" << endl;
	for (int i=0; i<OPT_CNT; ++i) {
		outstream << short_opts[i];
		outstream << "\t";
		outstream << long_opts[i];
		for (int tt=0; tt < 3 - int(long_opts[i].size())/8; ++tt)
			outstream << "\t";
		outstream << help_opts[i];
		outstream << "\n";
	}
	
	return;
}


int ACOptions::parse_options (int argc, char * argv[])
{
	for (int ar=1; ar<argc; ++ar) {
		int option = get_option_code(string(argv[ar]));
		
		switch (option) {
			case OPT_FILENAME:
				filename = string(argv[ar+1]);
				++ar;
				break;
			case OPT_SCORETYPE:
				scoretype = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_SEARCHSTRATEGY:
				searchstrategy = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_K:
				k = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_STOP:
				stop = true;
				break;
			case OPT_BEAMWIDTH:
				beamwidth = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_ITERATIONS:
				iterations = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_LAMBDA:
				lambda = atof(argv[ar+1]);
				++ar;
				break;
			case OPT_CYCLES:
				cycles = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_PARTITION:
				partitionfilename = string(argv[ar+1]);
				++ar;
				break;
			case OPT_OUTPUT:
				output = string(argv[ar+1]);
				++ar;
				break;
			case OPT_SCORES:
				scores = string(argv[ar+1]);
				++ar;
				break;
			case OPT_VERBOSE:
				verbosity = atoi(argv[ar+1]);
				++ar;
				break;
			case OPT_QUIET:
				verbosity = 0;
				break;				
			default:
				if (string(argv[ar]).operator [](0)=='-') {
					cerr << "Unknown option: " << argv[ar] << endl;
					return 1;
				}
				break;
		}
	}
	
	return 0;
}

