#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>
#include <vector>
#include <set>
#include <limits>
#include <time.h>
#include <mpfr.h>

#include "attributecluster.h"
#include "auxiliary.h"
#include "similaritymatrix.h"
#include "trie.h"
#include "bitvectormap.h"

#include "pugixml/src/pugixml.hpp"

using namespace std;
using namespace pugi;


const string AttrClusterer::name    = "Attribute Clustering";
const string AttrClusterer::version = "v2.0.20110317";
const string AttrClusterer::url     = "http://www.adrem.ua.ac.be/implementations";
const string AttrClusterer::author  = "Michael Mampaey";


map<int,string> AttrClusterer::scoretypenames = map<int,string>();
map<int,string> AttrClusterer::searchstrategynames = map<int,string>();



/* Constructor
 */
AttrClusterer::AttrClusterer()
{
	trans_count	= 0;
	attr_count	= 0;
	item_freqs	= 0;
	verbosity	= 1;
	
	datatype      = DATA_NONE;
	resultformat  = FMT_PLAINTXT;
	
	param_k          = 1;
	param_stop       = false;
	param_beam_width = 1;
	param_lambda     = 1.0;
	param_iterations = 1000;
	param_cycles     = 10;
	
	clusters             = 0;
	cluster_descr_length = 0;
	
	smallest_attr = 0;
	
	log_bell = numeric_limits<double>::quiet_NaN();
	
	running_time  = 0;
	
	scoretypenames[SCO_ENTROPY]     = "entropy";
	scoretypenames[SCO_TWOPART]     = "twopart-MDL";
	scoretypenames[SCO_SEMIREFINED] = "semi-refined";
	scoretypenames[SCO_REFINED]     = "refined";
	scoretypenames[SCO_BIC]         = "BIC";
	scoretypenames[SCO_AIC]         = "AIC";
	scoretypenames[SCO_COUNTDOWN]   = "countdown";
	scoretypenames[SCO_TEST]        = "test";
	
	searchstrategynames[SRCH_INPUT]        = "input";
	searchstrategynames[SRCH_HIERARCHICAL] = "hierarchical";
	searchstrategynames[SRCH_ANNEAL]       = "annealing";
	searchstrategynames[SRCH_BEAM]         = "beam-search";
	
	return;
}



/* Destructor
 */
AttrClusterer::~AttrClusterer()
{
	delete [] item_freqs;
	
	cleanup_clusters();
	
	for (vector<int*>::iterator it=binary_transactions.begin(); it!=binary_transactions.end(); ++it)
		delete [] *it;
	
	return;
}



/* Set verbosity level
 */
void AttrClusterer::set_verbosity(int v)
{
	if (v < 0)
		v = 0;
	verbosity = v;
	
	return;
}


/* Set score type
 */
void  AttrClusterer::set_scoretype(int st)
{
	scoretype = st;
	
	return;
}


/* Set search strategy
 */
void AttrClusterer::set_searchstrategy(int ss)
{
	searchstrategy = ss;
	
	return;
}



/* Set k
 */
void AttrClusterer::set_param_k(int k)
{
	param_k = k;
	
	return;
}


/* Set stop
 */
void AttrClusterer::set_param_stop(bool stop)
{
	param_stop = stop;
	
	return;
}


/* Set beam widht
 */
void AttrClusterer::set_param_beam_width(int w)
{
	if (w < 1)
		w = 1;
	param_beam_width = w;
	
	return;
}


/* Set lambda
 */
void AttrClusterer::set_param_lambda(double lambda)
{
	param_lambda = lambda;
	
	return;
}


/* Set iterations
 */
void AttrClusterer::set_param_iterations(int iter)
{
	if (iter < 0)
		iter = 0;
	param_iterations = iter;
	
	return;
}


/* Set cycles
 */
void AttrClusterer::set_param_cycles(int cycles)
{
	if (cycles < 1)
		cycles = 1;
	param_cycles = cycles;
	
	return;
}



/* Store filename in object and determine datatype based on extension
   returns true if ok
 */
bool AttrClusterer::set_data_filename(string filename)
{
	datafilename = filename;
	
	string extension;
	size_t pointpos = datafilename.find_last_of('.');
	if (pointpos == string::npos)
		extension = "";
	else
		extension = datafilename.substr(pointpos+1, datafilename.size()-pointpos-1);
	
	if (extension == "") {
		datatype = DATA_NONE;
	}
	else if (extension == "dat" || extension == "fimi") {
		datatype = DATA_BINARY;
		return true;
	}
	else if (extension == "data" || extension == "num" || extension == "csv") {
		datatype = DATA_CATEGORICAL;
		return true;
	}
	else {
		datatype = DATA_NONE;
	}
	
	if (verbosity >= 1)
		printf("invalid filename '%s'\n", datafilename.c_str());
	
	return false;
}



/* Read binary ".dat" or ".fimi", or categorical ".data" file
   returns true upon success
 */
bool AttrClusterer::read_data()
{
	if (datatype == DATA_NONE)
	{
		if (verbosity >= 1)
			printf("invalid filename '%s'\n", datafilename.c_str());
		
		return false;
	}
	else if (datatype == DATA_BINARY)
	{
		if (verbosity >= 1)
			printf("reading data file '%s'\n", datafilename.c_str());
		
		transactions = vector<Itemset>();
		attributes = set<int>();
		items = set<int>();
		map<int,double> item_freq_map = map<int,double>();
		
		FILE *f = fopen(datafilename.c_str(), "rt");
		
		int attr, item;
		while (!feof(f)) {
			string line;
			char c = getc(f);
			while (!feof(f) && c!= '\n' ) {
				line.push_back(c);
				c = getc(f);
			}
			
			if (line.size() && line[0] == '#') // comment line
				continue;
			if (line.size()==0 && feof(f)) 
				continue;
			if (line[line.size()-1]!=' ')
				line.push_back(' ');
			unsigned pos = 0, prev = 0;
			vector<int> trans = vector<int>();
			while (pos < line.size()) {
				prev = pos;
				pos = line.find(' ', prev+1);
				if (pos > prev && pos < line.size()) {
					attr = atoi(line.substr(prev, pos-prev).c_str());
					item = 2 * attr;
					trans.push_back(item);
				}
			}
			
			sort(trans.begin(), trans.end());
			Itemset t = Itemset(trans);
			
			for (vector<int>::iterator siit=trans.begin(); siit!=trans.end(); ++siit) {
				item = *siit;
				attr = item/2;
				attributes.insert(attr);
				items.insert(item);
				items.insert(item+1); // ?
				if (attr_items.find(attr)==attr_items.end()) {
					attr_items[attr] = set<int>();
					attr_items[attr].insert(item);
					attr_items[attr].insert(item+1); // dummy s.t. |{A=v}|=2
				}
				if (item_freq_map.find(item)==item_freq_map.end())
					item_freq_map[item] = 1;
				else
					item_freq_map[item] += 1;
			}
			
			transactions.push_back(t);
		}
		fclose(f);
		
		trans_count = transactions.size();
		attr_count = attributes.size();
		item_count = 2 * attr_count;
		
		smallest_attr = 1;
		if (attributes.find(0)!=attributes.end())
			smallest_attr = 0;
		
		item_freqs = new double[item_count + 2];
		
		for (map<int,double>::iterator midit=item_freq_map.begin(); midit!=item_freq_map.end(); ++midit) {
			item_freqs[midit->first]   = (double)midit->second / trans_count;
			item_freqs[(midit->first)+1] = (double)(trans_count - midit->second) / trans_count;
		}
		
	}
	else if (datatype == DATA_CATEGORICAL)
	{
		if (verbosity >= 1)
			printf("reading data file '%s'\n", datafilename.c_str());
		
		transactions = vector<Itemset>();
		attributes = set<int>();
		items = set<int>();
		map<int,double> item_freq_map = map<int,double>();
		
		FILE *f = fopen(datafilename.c_str(), "rt");
		
		attr_count = 0;
		item_count = 0;
		string value = "";
		
		while (!feof(f))
		{
			string line;
			char c = getc(f);
			while (!feof(f) && c!= '\n' ) {
				line.push_back(c);
				c = getc(f);
			}
			
			if (line.size() && line[0] == '#') // comment line
				continue;
			if (line.size()==0 && feof(f))
				continue;
			
			line.push_back(',');
			
			if (attr_count == 0) { // this is the first proper line containing data, count the attributes by commas
				for (unsigned i=0; i<line.size(); ++i) {
					if (line[i] == ',') {
						attr_items[attr_count] = set<int>();
						attributes.insert(attr_count);
						++attr_count;
					}
				}
			}
			
			unsigned pos = 0, prev = 0;
			vector<int> trans = vector<int>();
			int attr=0, item;
			pair<map<pair<int,string>, int>::iterator, bool> ret;
			while (pos < line.size()) {
				prev = pos;
				pos = line.find(',', prev+1);
				if (pos > prev && pos < line.size()) {
					value = line.substr(prev, pos-prev);
					// trim whitespace of value
					size_t first_char, last_char;
					first_char = value.find_first_not_of(" \t");
					last_char  = value.find_last_not_of(" \t");
					if (first_char==string::npos || last_char==string::npos)
						value = "";
					else if (first_char > 0 || last_char < value.size()-1)
						value = value.substr(first_char, last_char-first_char+1);
					
					ret = attrval_2_item.insert(pair<pair<int,string>, int>(pair<int,string>(attr,value), item_count));
					if (ret.second == false) {  // A=v was already in the map
						item = ret.first->second;
					}
					else {
						item = item_count;
						items.insert(item);
						attr_items[attr].insert(item);
						++item_count;
					}
					
					trans.push_back(item);
				}
				++attr;
			}
			
			sort(trans.begin(), trans.end());
			Itemset t = Itemset(trans);
			
			for (vector<int>::iterator siit=trans.begin(); siit!=trans.end(); ++siit) {
				item = *siit;
				if (item_freq_map.find(item)==item_freq_map.end())
					item_freq_map[item] = 1;
				else
					item_freq_map[item] += 1;
			}
			
			transactions.push_back(t);
			
		}
		
		fclose(f);
		
		trans_count = transactions.size();
		
		smallest_attr = 0;
		
		item_freqs = new double[item_count+1];
		
		for (map<int,double>::iterator midit=item_freq_map.begin(); midit!=item_freq_map.end(); ++midit) {
			item_freqs[midit->first] = midit->second / trans_count;
		}
		
	}
	
	transactions_to_binary();
	
	if (verbosity >= 1) {
		printf("%i transactions, %i attributes\n", trans_count, attr_count);
		printf("canonical description length: %.2f\n\n", canonical_descrlen());
	}
	
	return true;
}



/* horizontal to vertical format
 */
void AttrClusterer::transactions_to_binary()
{
	vector<Itemset>::iterator vit;
	for (vit=transactions.begin(); vit!=transactions.end(); ++vit) {
		int * b = new int[item_count+2];
		memset(b, 0, sizeof(int) * (item_count+2));
		if (datatype == DATA_CATEGORICAL) {
			for (int i=0; i<vit->size; ++i)
				b[vit->iset[i]] = 1;
		}
		else if (datatype == DATA_BINARY) {
			for (int i=0; i<item_count+2; i+=2)
				if (vit->contains(i))
					b[i] = 1;
				else
					b[i+1] = 1;
		}
		binary_transactions.push_back(b);
	}
	
	transactions.clear();
	
	return;
}



/* Initialize the clustering(s)
 */
void AttrClusterer::init_clusters()
{
	if (verbosity >= 1)
		printf("initializing clusters\n");
	
	clusters = new map<int, Attrset>[param_beam_width];
	cluster_descr_length = new map<int, pair<double,double> >[param_beam_width];
	
	for (int w=0; w<param_beam_width; ++w) {
		clusters[w] = map<int, Attrset>();
		cluster_descr_length[w] = map<int, pair<double,double> >();
		
		int k = 0;
		for (set<int>::iterator siit=attributes.begin(); siit!=attributes.end(); ++siit) {
			int attribute = *siit;
			clusters[w].insert(pair<int,Attrset>(k, Attrset(attribute)));
			cluster_descr_length[w][k] = singleton_descr_len(attribute);
			++k;
		}
	}
	
	return;
}



/* Initialize random clustering(s)
   Note: not uniform over all clusterings, but clusterings of the same size are generated w equal probability
 */
void AttrClusterer::init_clusters_random(bool quiet)
{
	if (!quiet && verbosity >= 1)
		printf("initializing clusters at random\n");
	
	srand(time(0));
	
	clusters = new map<int, Attrset>[param_beam_width];
	cluster_descr_length = new map<int, pair<double,double> >[param_beam_width];
	
	for (int w=0; w<param_beam_width; ++w) {
		clusters[w] = map<int, Attrset>();
		cluster_descr_length[w] = map<int, pair<double,double> >();
		
		map<int,int> label_map = map<int,int>();
		map<int,vector<int> > tmp_clusters;
		int cluster_count = 0;
		for (set<int>::iterator siit=attributes.begin(); siit!=attributes.end(); ++siit) {
			int label = randint(0, attr_count);
			if (label_map.find(label)==label_map.end()) {
				label_map[label] = cluster_count;
				tmp_clusters[cluster_count] = vector<int>();
				++cluster_count;
			}
			tmp_clusters[label_map[label]].push_back(*siit);
		}
		
		for (map<int,vector<int> >::iterator mivit=tmp_clusters.begin(); mivit!=tmp_clusters.end(); ++mivit) {
			clusters[w].insert(pair<int,Attrset>(mivit->first, Attrset(mivit->second)));
			cluster_descr_length[w][mivit->first] = cluster_descr_len(mivit->first, w);
		}
	}
	
	return;
}



/* Cleanup
 */
void AttrClusterer::cleanup_clusters()
{
	delete [] clusters;
	delete [] cluster_descr_length;
	
	clusters             = 0;
	cluster_descr_length = 0;
	
	return;
}



/* Compute gain if clusters are merged.
   positive if merging i & j is a good merge
   omits difference in partition description.
 */
double AttrClusterer::cluster_merge_gain(int i, int j, int n)
{
	if (i > j) swap(i, j);
	double sim = 0.0;
	pair<double,double> sc = attrset_descr_len(i, j, n);
	sim += sc.first + sc.second;
	sc = cluster_descr_length[n][i];
	sim -= sc.first + sc.second;
	sc = cluster_descr_length[n][j];
	sim -= sc.first + sc.second;

	return - sim;
}



/* Merge clusters i and j, store into cluster w smallest ID, and erase other one
   n is the clustering identifier (only for beam search, otherwise n=0)
 */
void AttrClusterer::merge_clusters(int i, int j, int n)
{
	if (i == j) return;
	if (i > j) swap(i, j);

	Attrset as = Attrset(clusters[n][i], clusters[n][j]);
	clusters[n].erase(i);
	clusters[n].insert(pair<int,Attrset>(i, as));
	cluster_descr_length[n][i] = cluster_descr_len(i, n);
	clusters[n].erase(j);
	cluster_descr_length[n].erase(j);

	return;
}



/*
 */
void AttrClusterer::split_cluster(int i, const Attrset & aset1, const Attrset & aset2, int n)
{
	int cl1 = i;
	int cl2 = 0;
	while (clusters[n].find(cl2)!=clusters[n].end()) ++cl2;
	clusters[n].erase(i);
	clusters[n].insert(pair<int,Attrset>(cl1, aset1));
	cluster_descr_length[n][cl1] = cluster_descr_len(cl1);
	clusters[n].insert(pair<int,Attrset>(cl2, aset2));
	cluster_descr_length[n][cl2] = cluster_descr_len(cl2);
	
	return;
}



/* Compute description lenght of a single attribute
 */
pair<double,double> AttrClusterer::singleton_descr_len(int attr)
{
	// data description length
	double data_score = 0.0;
	double total_freq = 0.0;
	
	double f = 0.0;
	
	int val_cnt = attr_items[attr].size();
	
	set<int>::iterator siit = attr_items[attr].begin();
	while (siit!=attr_items[attr].end() && (1.0-total_freq) * trans_count >= 1.0)
	{
		if (siit != attr_items[attr].end())
			f = item_freqs[*siit];
		else
			f = 1.0 - total_freq;
		
		switch (scoretype) {
			case SCO_ENTROPY:
			case SCO_TWOPART:
			case SCO_BIC:
			case SCO_AIC:
				data_score -= f * log_2(f) * trans_count;
				break;
			case SCO_SEMIREFINED:
			case SCO_REFINED:
			case SCO_TEST:
				data_score -= log_factorial(2*f*trans_count) - log_factorial(f*trans_count) - 2*f*trans_count;
				break;
			case SCO_COUNTDOWN:
				data_score -= log_factorial(f*trans_count);
				break;
			default:
				break;
		}
		if (siit != attr_items[attr].end())
			++siit;
		total_freq += f;
	}
	
	switch (scoretype) {
		case SCO_SEMIREFINED: // domsize==valcount for singletons
		case SCO_REFINED:
		case SCO_TEST:
			if (val_cnt % 2 == 1)
				data_score += log_factorial(2*trans_count + (val_cnt-1)) - log_factorial(trans_count + val_cnt/2) - 2*trans_count - log_factorial(val_cnt-1) + log_factorial(val_cnt/2);
			else
				data_score += log_factorial(trans_count + val_cnt/2 - 1) - log_factorial(val_cnt/2 - 1);
			break;
		case SCO_COUNTDOWN:
			data_score += log_factorial(trans_count);
			break;
		case SCO_ENTROPY:
		case SCO_TWOPART:
		case SCO_BIC:
		case SCO_AIC:
			// nothing
			break;
		default:
			break;
	}
	
	// model description length (code table)
	double model_score = 0.0;
	
	switch (scoretype) {
		case SCO_ENTROPY:
		case SCO_REFINED:
			model_score = 0.0;
			break;
		case SCO_SEMIREFINED:
			model_score = (val_cnt <= std::numeric_limits<double>::digits) ? log_2(pow(2, val_cnt) - 1) : val_cnt;
			break;
		case SCO_TWOPART:
			// for singletons domainsize==val_cnt 
			total_freq = 0.0;
			siit = attr_items[attr].begin();
			while (siit!=attr_items[attr].end() && (1.0-total_freq) * trans_count >= 1.0)
			{
				if (siit != attr_items[attr].end())
					f = item_freqs[*siit];
				else
					f = 1.0 - total_freq;
				model_score += log_2(val_cnt) + log_2(log_2(trans_count)) - log_2(f);
				if (siit != attr_items[attr].end())
					++siit;
				total_freq += f;
			}
			break;
		case SCO_BIC:
			model_score = (val_cnt - 1) * log_2(trans_count) / 2.0;
			break;
		case SCO_AIC:
			model_score = val_cnt - 1;
			break;
		case SCO_COUNTDOWN:
			model_score = val_cnt * (1 + log_2(trans_count));
			break;
		case SCO_TEST:
			model_score = log_2(val_cnt) + val_cnt * log_2(log_2(trans_count));
			break;
		default:
			break;
	}
	
	return pair<double,double>(model_score, data_score);
}



pair<double,double> AttrClusterer::cluster_descr_len(int i, int n)
{
	return attrset_descr_len(i, i, n);
}



pair<double,double> AttrClusterer::attrset_descr_len(int i, int j, int n)
{
	if (i > j) swap(i, j);
	
	Attrset & attrset1 = clusters[n][i];
	Attrset & attrset2 = clusters[n][j];
	
	return attrset_descr_len(attrset1, attrset2);
}



/* computes the description length if they were merged (?)
   attrsets must either be equal or disjoint
 */
pair<double,double> AttrClusterer::attrset_descr_len(const Attrset & attrset1, const Attrset & attrset2)
{
	bool equal = (attrset1 == attrset2);
	
	const int attrset_size = equal ? attrset1.size : attrset1.size + attrset2.size;
	
	Attrset tmpset = equal ? attrset1 : Attrset(attrset1, attrset2);
	// note: partition.size() <= min(2**itemset.size(), trans_count)
	
	int itemunionsize = 0;
	
	for (int i=0; i< tmpset.size; ++i)
		itemunionsize += attr_items[tmpset[i]].size() - 1; // we can omit one attribute-value
	
	int * itemunion = new int[itemunionsize];
	
	itemunionsize = 0;
	for (int i=0; i<tmpset.size; ++i) {
		set<int>::iterator siit=attr_items[tmpset[i]].begin();
		++siit; // we can omit one attribute-value
		while (siit!=attr_items[tmpset[i]].end()) {
			itemunion[itemunionsize] = *siit;
			++siit;
			++itemunionsize;
		}
	}
	
	sort(itemunion, itemunion+itemunionsize); // necessary?
	
	BitVectorMap bvm(itemunionsize);
	int * value = new int[itemunionsize];
	
	for (vector<int*>::iterator it=binary_transactions.begin(); it!=binary_transactions.end(); ++it) {
		for (int i=0; i<itemunionsize; ++i)
			value[i] = (*it)[itemunion[i]];
		bvm.increment(value);
	}
	
	delete [] itemunion;
	delete [] value;
	
	pair<double,double> hls = bvm.get_entropy_and_log_sum();
	
	double domainsize = 1.0;
	for (int i=0; i<tmpset.size; ++i)
		domainsize *= attr_items[tmpset[i]].size();
	
	// data description length;
	double data_score = 0.0;
	
	double entropy = hls.first;
	switch (scoretype) {
		case SCO_ENTROPY:
		case SCO_TWOPART:
		case SCO_BIC:
		case SCO_AIC:
			data_score = entropy * trans_count;
			break;
		case SCO_SEMIREFINED:
		case SCO_TEST:
			data_score = bvm.get_refined_entropy(bvm.get_size());
			break;
		case SCO_REFINED:
			data_score = bvm.get_refined_entropy(domainsize);
			break;
		case SCO_COUNTDOWN:
			data_score = log_factorial(trans_count) - bvm.map_sum(&log_factorial<double>); // template function pointers, how do they work?
			break;
		default:
			break;
			
	}
	
	// model description length (code table)
	double model_score = 0.0;
	
	switch (scoretype) {
		case SCO_ENTROPY:
		case SCO_REFINED:
			model_score = 0.0;
			break;
		case SCO_SEMIREFINED:
			model_score = (domainsize <= std::numeric_limits<double>::digits) ? log_2(pow(2, domainsize) - 1) : domainsize;
			break;
		case SCO_TWOPART:
			model_score = bvm.get_size() * (log_2(domainsize) + log_2(log_2(trans_count))) + hls.second;
			break;
		case SCO_BIC:
			model_score = (domainsize - 1) * log_2(trans_count) / 2.0;
			break;
		case SCO_AIC:
			model_score = domainsize - 1;
			break;
		case SCO_COUNTDOWN:
			model_score = bvm.get_size() * (attrset_size + log_2(trans_count));
			break;
		case SCO_TEST:
			model_score = log_2(domainsize) + log_factorial(domainsize) - log_factorial(bvm.get_size()) - log_factorial(domainsize-bvm.get_size()) + bvm.get_size() * log_2(log_2(trans_count));
			break;
		default:
			break;
	}
	
	return pair<double,double>(model_score, data_score);
}



/* Get (compute) MDL score (of clustering in beam n)
 */
triplet<double> AttrClusterer::clustering_score(int n)
{
	
	double model_score = 0.0;
	double data_score  = 0.0;
	
	model_score += partition_descr_len();
	
	map<int,Attrset>::iterator misiit;
	for (misiit=clusters[n].begin(); misiit!=clusters[n].end(); ++misiit) {
		int i = misiit->first;
		model_score += cluster_descr_length[n][i].first;
		data_score += cluster_descr_length[n][i].second;
	}
	
	return triplet<double>(model_score+data_score, model_score, data_score);
}



/* Description length of the attribute partition (sans the distribution descriptions)
 */
double AttrClusterer::partition_descr_len()
{	
	if (isnan(log_bell))
		log_bell = log_bell_number(attr_count);
	
	return log_bell;
}



/* Compute the MDL score of a partition given in 'partitionfilename'
 */
triplet<double> AttrClusterer::input_descrlen(string partitionfilename)
{
	
	if (verbosity >= 1) {
		printf("Reading and computing score of partition in '%s' ...\n", partitionfilename.c_str());
	}
	
	cleanup_clusters();
	clusters = new map<int, Attrset>[1];
	cluster_descr_length = new map<int, pair<double,double> >[1];
	clusters[0] = map<int, Attrset>();
	cluster_descr_length[0] = map<int, pair<double,double> >();
	
	int k = 0;
	
	resultformat = FMT_PLAINTXT;
	
	if (resultformat == FMT_PLAINTXT) {
		FILE *f = fopen(partitionfilename.c_str(), "rt");
		
		int item;
		k = 0;
		while (!feof(f)) {
			string line;
			char c = getc(f);
			while (!feof(f) && c!= '\n' ) {
				line.push_back(c);
				c = getc(f);
			}
			
			if (line.size() && line[0] == '#') continue;
			if (line.size()==0 && feof(f)) continue;
			if (line[line.size()-1]!=' ')
				line.push_back(' ');
			unsigned pos = 0, prev = 0;
			vector<int> trans = vector<int>();
			while (pos < line.size()) {
				prev = pos;
				pos = line.find(' ', prev+1);
				if (pos > prev && pos < line.size()) {
					item = atoi(line.substr(prev, pos-prev).c_str());
					trans.push_back(item);
				}
			}
			
			Attrset t = Attrset(trans);
			clusters[0].insert(pair<int,Attrset>(k, t));
			cluster_descr_length[0][k] = attrset_descr_len(k, k);
			++k;
			
		}
		fclose(f);
	}
	else if (resultformat == FMT_XML) {
		xml_document doc;
		
		doc.load_file(partitionfilename.c_str());
		
		// DO STUFF
		
	}
	
	double model_score = 0.0;
	double data_score = 0.0;
	
	for (int i=0; i<k; ++i) {
		model_score += cluster_descr_length[0][i].first;
		data_score += cluster_descr_length[0][i].second;
	}
	
	// partition description
	model_score += partition_descr_len();
	
	overall_best_clustering = map<int, Attrset >(clusters[0]);
	overall_best_clustering_score = triplet<double>(model_score+data_score, model_score, data_score);
	
	if (verbosity >= 1) {
		printf("%.2f\t%.2f\t%.2f\n", overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third);
	}
	
	return overall_best_clustering_score;
}



/* Canonical description length of data -- wrt uniform prior
 */
double AttrClusterer::canonical_descrlen()
{
	double descr_len = 0.0;
	
	for (map<int, set<int> >::iterator mit=attr_items.begin(); mit!=attr_items.end(); ++mit) {
		descr_len += log_2(mit->second.size());
	}
	
	return descr_len * trans_count;
	
}



/*
 */
void AttrClusterer::hierarchical_agglomerative()
{
	if (param_k == 0)
		param_k = 1;
	
	set_param_beam_width(1);
	
	if (verbosity >= 1)
		printf("starting agglomerative clustering\n\n");
	
	time_t starttime = time(0);
	
	init_clusters();
	
	FILE *sf = 0;
	
	overall_best_clustering = map<int,Attrset>(clusters[0]);
	overall_best_clustering_score = clustering_score();
	
	if (verbosity >= 2) {
		printf("|A|\tdelta\t\tL(A,D)\t   L(A)\t   L(D|A)\n");
		printf("%d\t0.0\t\t", int(clusters[0].size()));
		printf("%.2f = %.2f + %.2f\n", overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third);
	}
	if (scorefilename.size()) {
		sf = fopen(scorefilename.c_str(), "w");
		fprintf(sf, "%d\t%f\t%f\t%f\n" , int(clusters[0].size()), overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third);
	}
	
	// initial pairwise (singleton) similarity computation
	SimilarityMatrix cluster_similarities = SimilarityMatrix(attr_count);
	
	for (int i=0; i<attr_count; ++i) {
		for (int j=i+1; j<attr_count; ++j) {
			cluster_similarities.insert(i, j, cluster_merge_gain(i, j));
		}
	}
	
	double prev_descr_len = numeric_limits<double>::infinity();
	double curr_descr_len = overall_best_clustering_score.first;
	
	while (int(clusters[0].size()) > param_k && (!param_stop || prev_descr_len > curr_descr_len))
	{
		// determine the best merge
		pair<double, pair<int,int> > bestest = cluster_similarities.get_max();
		double bestscore = bestest.first;
		pair<int,int> bestpair = bestest.second;
		
		// perform merge
		merge_clusters(bestpair.first, bestpair.second);
		cluster_similarities.delete_row(bestpair.first);
		cluster_similarities.delete_row(bestpair.second);
		cluster_similarities.init_row(bestpair.first);
		
		// compute the new cluster similarities
		map<int, Attrset>::iterator miiit;
		for (miiit=clusters[0].begin(); miiit!=clusters[0].end(); ++miiit) {
			int i = miiit->first;
			if (i == bestpair.first)
				continue;
			cluster_similarities.insert(i, bestpair.first, cluster_merge_gain(i, bestpair.first));
		}
		
		// current score
		prev_descr_len = curr_descr_len;
		triplet<double> curr_mdl_score = clustering_score();
		curr_descr_len = curr_mdl_score.first;
		
		// keep track of best clustering seen so far
		if (curr_mdl_score.first < overall_best_clustering_score.first) {
			overall_best_clustering.clear();
			overall_best_clustering = map<int, Attrset >(clusters[0]);
			overall_best_clustering_score = curr_mdl_score;
		}
		
		if (verbosity >= 2) {
			printf("%d\t%.2f\t\t", int(clusters[0].size()), bestscore);
			printf("%.2f = %.2f + %.2f\n", curr_mdl_score.first, curr_mdl_score.second, curr_mdl_score.third);
		}
		if (scorefilename.size())
			fprintf(sf, "%d\t%f\t%f\t%f\n", int(clusters[0].size()), curr_mdl_score.first, curr_mdl_score.second, curr_mdl_score.third);
	}
	
	if (scorefilename.size())
		fclose(sf);
	
	running_time = time(0) - starttime;
	
	if (verbosity >= 1)
		print_overall_best_clustering();
	
	return;
}



/* Simulated annealing
 */
void AttrClusterer::simulated_annealing()
{
	set_param_beam_width(1);
	
	if (verbosity >= 1)
		printf("starting simulated annealing\n\n");
	
	time_t starttime = time(0);
	
	double max_neighbor_gain = 0.0;
	
	map<int, Attrset>::iterator it;
	
	mpfr_t split_cnt, clustersize, randomsplit;
	mpfr_init2(split_cnt, attr_count);
	mpfr_init2(clustersize, attr_count);
	mpfr_init2(randomsplit, attr_count);
	gmp_randstate_t rnd_state;
	gmp_randinit_default (rnd_state);
	
	// get a rough estimte of the maximal score gain/loss between any two neighboring partitions
	for (int i=0; i<100; ++i)
	{
		init_clusters_random(true);
		int k = clusters[0].size();
		if (k==attr_count || (k > 1 && randfloat() < 0.5)) // do a merge
		{
			int cl1 = randint(0, k);
			int cl2 = randint(0, k-1);
			if (cl2 >= cl1) ++cl2;
			it = clusters[0].begin(); for (int i=0; i<cl1; ++i) ++it; cl1 = it->first;
			it = clusters[0].begin(); for (int i=0; i<cl2; ++i) ++it; cl2 = it->first;
			
			double gain = cluster_merge_gain(cl1, cl2);
			
			max_neighbor_gain = max(max_neighbor_gain, fabs(gain));
		}
		else // do a split
		{
			// count how many splits are possible
			mpfr_set_ui(split_cnt, 0, GMP_RNDN);
			for (it=clusters[0].begin(); it!=clusters[0].end(); ++it) {
				if (it->second.size > 1) {
					mpfr_set_ui(clustersize, 1, GMP_RNDN);
					mpfr_mul_2ui(clustersize, clustersize, it->second.size-1, GMP_RNDN);
					mpfr_sub_ui(clustersize, clustersize, 1, GMP_RNDN);
					mpfr_add(split_cnt, split_cnt, clustersize, GMP_RNDN);
				}
			}
			// pick a cluster to split, s.t. we pick splits uniformly
			mpfr_urandomb(randomsplit, rnd_state);
			it = clusters[0].begin();
			int cl = -1;
			while (it!=clusters[0].end()) {
				if (it->second.size > 1) {
					cl = it->first;
					mpfr_set_ui(clustersize, 1, GMP_RNDN);
					mpfr_mul_2ui(clustersize, clustersize, it->second.size-1, GMP_RNDN);
					mpfr_sub_ui(clustersize, clustersize, 1, GMP_RNDN);
					mpfr_div(clustersize, clustersize, split_cnt, GMP_RNDN);
					mpfr_sub(randomsplit, randomsplit, clustersize, GMP_RNDN);
					if (mpfr_sgn(randomsplit) < 0) {
						break;
					}
				}
				++it;
			}
			
			// divide the chosen cluster, in order to compute the score difference
			vector<int> set1, set2;
			while (set1.size()==0 || set2.size()==0) { // make sure we do a proper split
				set1.clear();
				set2.clear();
				for (int i=0; i<clusters[0][cl].size; ++i) {
					if (randfloat() < 0.5)
						set1.push_back(clusters[0][cl][i]);
					else
						set2.push_back(clusters[0][cl][i]);
				}
				sort(set1.begin(), set1.end());
				sort(set2.begin(), set2.end());
			}
			
			Attrset aset1 = Attrset(set1);
			Attrset aset2 = Attrset(set2);
			pair<double,double> set1_cdl = attrset_descr_len(aset1, aset1);
			pair<double,double> set2_cdl = attrset_descr_len(aset2, aset2);
			pair<double,double> cl_cdl = cluster_descr_length[0][cl];
			
			double gain = (cl_cdl.first+cl_cdl.second) - (set1_cdl.first+set1_cdl.second) - (set2_cdl.first+set2_cdl.second);
			
			max_neighbor_gain = max(max_neighbor_gain, fabs(gain));
		}
		
		cleanup_clusters();
	}
	
	max_neighbor_gain *= 2; // heuristically
	if (verbosity >= 3)
		printf("max_neighbor_gain: %f\n", max_neighbor_gain);
	
	triplet<double> curr_score = triplet<double>(numeric_limits<double>::infinity(), numeric_limits<double>::infinity(), numeric_limits<double>::infinity());
	overall_best_clustering.clear();
	overall_best_clustering_score = curr_score;
	
	//choose temperature scaling factor alpha such that after 'steps' steps, (double)alpha^steps==0.0 due to rounding
	// therefore, the nr of steps directly affects the cooling rate
	double alpha = pow(2.0, (numeric_limits<double>::min_exponent + log(numeric_limits<double>::epsilon())/log(2)) / (double)param_iterations);
	//double lambda = 1.0 / (-log(0.50) * pow(alpha, steps/2));
	
	
	for (int cycle=0; cycle<param_cycles; ++cycle)
	{
		if (verbosity >=2 )
			printf("CYCLE %d\n", cycle);
		
		srand(time(0));
		
		cleanup_clusters();
		init_clusters_random();
		
		int step = 0;
		double temperature = 1.0; // pow(alpha, step); // geometrically decrease temperature from 1.0 to 0.0
		while (temperature > 0.0 && step < param_iterations)
		{
			if (verbosity >= 2) {
				if (step % int(param_iterations/50.0) == 0) 
					printf("%i\t%u\t%f\n", step, (unsigned)clusters[0].size(), curr_score.first);
			}
			
			temperature = pow(alpha, step);
			
			int k = clusters[0].size();
			bool changed = false;
			// 50-50 merge or split (unless only one is possible)
			if (k==attr_count || (k > 1 && randfloat() < 0.5)) { // merge
				int cl1 = randint(0, k);
				int cl2 = randint(0, k-1);
				if (cl2 >= cl1) ++cl2;
				it = clusters[0].begin(); for (int i=0; i<cl1; ++i) ++it; cl1 = it->first;
				it = clusters[0].begin(); for (int i=0; i<cl2; ++i) ++it; cl2 = it->first;
				
				double gain = cluster_merge_gain(cl1, cl2);
				if (exp((gain/max_neighbor_gain) / (param_lambda*temperature)) > randfloat()) {
					merge_clusters(cl1, cl2);
					changed = true;
				}
			}
			else  // split
			{
				// count how many splits are possible
				mpfr_set_ui(split_cnt, 0, GMP_RNDN);
				for (it=clusters[0].begin(); it!=clusters[0].end(); ++it) {
					if (it->second.size > 1) {
						mpfr_set_ui(clustersize, 1, GMP_RNDN);
						mpfr_mul_2ui(clustersize, clustersize, it->second.size-1, GMP_RNDN);
						mpfr_sub_ui(clustersize, clustersize, 1, GMP_RNDN);
						mpfr_add(split_cnt, split_cnt, clustersize, GMP_RNDN);
					}
				}
				// pick a cluster to split, s.t. we pick splits uniformly
				mpfr_urandomb(randomsplit, rnd_state);
				it = clusters[0].begin();
				int cl = -1;
				while (it!=clusters[0].end()) {
					if (it->second.size > 1) {
						cl = it->first;
						mpfr_set_ui(clustersize, 1, GMP_RNDN);
						mpfr_mul_2ui(clustersize, clustersize, it->second.size-1, GMP_RNDN);
						mpfr_sub_ui(clustersize, clustersize, 1, GMP_RNDN);
						mpfr_div(clustersize, clustersize, split_cnt, GMP_RNDN);
						mpfr_sub(randomsplit, randomsplit, clustersize, GMP_RNDN);
						if (mpfr_sgn(randomsplit) < 0) { // split this one
							break;
						}
					}
					++it;
				}
				
				// divide the chosen cluster
				vector<int> set1, set2;
				while (set1.size()==0 || set2.size()==0) { // make sure we do a proper split
					set1.clear();
					set2.clear();
					for (int i=0; i<clusters[0][cl].size; ++i) {
						if (randfloat() < 0.5)
							set1.push_back(clusters[0][cl][i]);
						else
							set2.push_back(clusters[0][cl][i]);
					}
					sort(set1.begin(), set1.end());
					sort(set2.begin(), set2.end());
				}
				
				Attrset aset1 = Attrset(set1);
				Attrset aset2 = Attrset(set2);
				pair<double,double> set1_cdl = attrset_descr_len(aset1, aset1);
				pair<double,double> set2_cdl = attrset_descr_len(aset2, aset2);
				pair<double,double> cl_cdl = cluster_descr_length[0][cl];
				
				double gain = (cl_cdl.first+cl_cdl.second) - (set1_cdl.first+set1_cdl.second) - (set2_cdl.first+set2_cdl.second);
				
				if (exp((gain/max_neighbor_gain) / (param_lambda*temperature)) > randfloat())
				{
					// make the two new clusters
					split_cluster(cl, aset1, aset2, 0);
					changed = true;
				}
			}
			
			if (changed) {
				curr_score = clustering_score();
				if (overall_best_clustering_score.first > curr_score.first) {
					overall_best_clustering.clear();
					overall_best_clustering = map<int, Attrset >(clusters[0]);
					overall_best_clustering_score = curr_score;
				}
			}
			
			++step;
		}
	}
	
	mpfr_clear(split_cnt);
	mpfr_clear(clustersize);
	mpfr_clear(randomsplit);
	gmp_randclear(rnd_state);
	
	running_time = time(0) - starttime;
	
	if (verbosity >= 1)
		print_overall_best_clustering();
	
	return;
}



/* Beam search (hierarchical agglomerative) with given 'width', terminate at 'k' or 'stop' at local minimum
 */
void AttrClusterer::beam_search()
{
	set_param_beam_width(param_beam_width);
	
	if (verbosity >= 1)
		printf("starting beam search\n\n");
	
	time_t starttime = time(0);
	
	init_clusters();
	
	const int large_beam_width = 4 * param_beam_width; // to be absolutely safe this should be beam_widht**2 ...
	
	double       * top_scores     = new double[large_beam_width];
	triplet<int> * top_merges     = new triplet<int>[large_beam_width];
	
	overall_best_clustering       = map<int,Attrset>(clusters[0]);
	overall_best_clustering_score = clustering_score(0);
	
	int beam_len = 1; // <= beam_width
	
	// initial pairwise (singleton) similarity computation
	SimilarityMatrix ** cluster_similarities = new SimilarityMatrix*[beam_len];
	for (int w=0; w<beam_len; ++w)
		cluster_similarities[w] = new SimilarityMatrix(attr_count);
	
	for (int i=0; i<attr_count; ++i) {
		for (int j=i+1; j<attr_count; ++j) {
			double cmg = cluster_merge_gain(i, j, 0);
			cluster_similarities[0]->insert(i, j, cmg);
		}
	}
	
	double prev_best_descr_len = numeric_limits<double>::infinity();
	double curr_best_descr_len = overall_best_clustering_score.first;
	
	while (int(clusters[0].size()) > param_k && (!param_stop || prev_best_descr_len > curr_best_descr_len))
	{
		// find the top-'large_beam_width' best merges from the current beam; large_beam_width may be larger than beam_width
		pair<double, triplet<int> > * new_beam = new pair<double, triplet<int> >[large_beam_width];
		for (int w=0; w<large_beam_width; ++w)
			new_beam[w].first = - std::numeric_limits<double>::infinity();
		int large_beam_len = 0;
		for (int w=0; w<beam_len; ++w)
		{
			int local_toplen = 0;
			pair<double, pair<int,int> > *local_topk = new pair<double, pair<int,int> >[param_beam_width];
			cluster_similarities[w]->get_topk(param_beam_width, 0, local_topk, local_toplen);
			double mdl_w = clustering_score(w).first;
			for (int i=0; i<local_toplen; ++i) { // fill new_beam up to large_beam_width
				if (large_beam_len < large_beam_width) {
					new_beam[large_beam_len] = pair<double, triplet<int> >(mdl_w - local_topk[i].first, triplet<int>(w, local_topk[i].second.first, local_topk[i].second.second));
					++large_beam_len;
					std::sort(new_beam, new_beam+large_beam_len);
				}
				else if (local_topk[i].first > new_beam[large_beam_width-1].first) {
					new_beam[large_beam_width-1] = pair<double, triplet<int> >(mdl_w - local_topk[i].first, triplet<int>(w, local_topk[i].second.first, local_topk[i].second.second));
					std::sort(new_beam, new_beam+large_beam_width);
				}
			}
			delete [] local_topk;
		}
		
		for (int w=0; w<large_beam_len; ++w) {
			top_scores[w] = new_beam[w].first;
			top_merges[w] = triplet<int>(new_beam[w].second.first, new_beam[w].second.second, new_beam[w].second.third);
		}
		delete [] new_beam;
		
		
		// do it well
		SimilarityMatrix ** new_cluster_similarities = new SimilarityMatrix*[min(param_beam_width, large_beam_len)];
		map<int, Attrset> * new_clusters = new map<int, Attrset>[min(param_beam_width, large_beam_len)];
		map<int, pair<double,double> > * new_cluster_descr_length = new map<int, pair<double,double> >[min(param_beam_width, large_beam_len)];
		
		// construct the new beam, taking care of possible duplicates
		int new_beam_len = 0;
		for (int w=0; w<large_beam_len && new_beam_len<param_beam_width; ++w)
		{
			int source_cl = top_merges[w].first;
			new_cluster_similarities[new_beam_len] = new SimilarityMatrix(*(cluster_similarities[source_cl]));
			new_clusters[new_beam_len].clear();
			new_clusters[new_beam_len] = map<int, Attrset>(clusters[source_cl]);
			new_cluster_descr_length[new_beam_len] = map<int, pair<double,double> >(cluster_descr_length[source_cl]);
			
			swap(clusters, new_clusters);
			swap(cluster_descr_length, new_cluster_descr_length);
			
			// merge
			int i = top_merges[w].second;
			int j = top_merges[w].third;
			if (i > j) swap(i, j);
			merge_clusters(i, j, new_beam_len);
			new_cluster_similarities[new_beam_len]->delete_row(i);
			new_cluster_similarities[new_beam_len]->delete_row(j);
			new_cluster_similarities[new_beam_len]->init_row(i);
			
			map<int, Attrset>:: iterator miiit;
			for (miiit=clusters[new_beam_len].begin(); miiit!=clusters[new_beam_len].end(); ++miiit) {
				int ii = miiit->first;
				if (ii == i)
					continue;
				new_cluster_similarities[new_beam_len]->insert(ii, i, cluster_merge_gain(ii, i, new_beam_len));
			}
			
			swap(clusters, new_clusters);
			swap(cluster_descr_length, new_cluster_descr_length);
			
			bool is_duplicate = false;
			for (int ww=0; ww<new_beam_len; ++ww) {
				if (new_clusters[ww] == new_clusters[new_beam_len]) {
					new_clusters[new_beam_len].clear();
					new_cluster_descr_length[new_beam_len].clear();
					delete new_cluster_similarities[new_beam_len];
					new_cluster_similarities[new_beam_len] = 0;
					is_duplicate = true;
					break;
				}
			}
			
			if (!is_duplicate)
				++new_beam_len;
			
		}
		
		// out with the old, in with the new
		for (int w=0; w<beam_len; ++w)
			delete cluster_similarities[w];
		delete [] cluster_similarities;
		cluster_similarities = new_cluster_similarities;
		cleanup_clusters();
		clusters = new_clusters;
		cluster_descr_length = new_cluster_descr_length;
		
		beam_len = new_beam_len;
		
		
		// update best score yet found
		// should be redundant
		int curr_best_w = 0;
		prev_best_descr_len = curr_best_descr_len;
		triplet<double> curr_mdl_score = clustering_score(0);
		curr_best_descr_len = curr_mdl_score.first;
		for (int w=1; w<beam_len; ++w) {
			triplet<double> mw = clustering_score(w);
			if (mw.first < curr_best_descr_len) {
				curr_best_w = w;
				curr_mdl_score = mw;
				curr_best_descr_len = curr_mdl_score.first;
			}
		}
		
		// keep track of best clustering seen so far
		if (curr_mdl_score.first < overall_best_clustering_score.first) {
			overall_best_clustering.clear();
			overall_best_clustering = map<int, Attrset >(clusters[curr_best_w]);
			overall_best_clustering_score = curr_mdl_score;
		}
		
		if (verbosity >= 2) {
			printf("%d\t%.2f\t\t", int(clusters[curr_best_w].size()), overall_best_clustering_score.first);
			printf("%.2f = %.2f + %.2f\n", curr_mdl_score.first, curr_mdl_score.second, curr_mdl_score.third);
			//printf("%i\t%i\n", beam_len, curr_best_w);
		}
		
	}
	
	for (int w=0; w<beam_len; ++w)
		delete cluster_similarities[w];
	delete [] cluster_similarities;
	
	delete [] top_scores;
	delete [] top_merges;
	
	running_time = time(0) - starttime;
	
	if (verbosity >= 1)
		print_overall_best_clustering();
	
	return;
}



/* Print the 'overall_best_clustering'
 */
void AttrClusterer::print_overall_best_clustering()
{
	//triplet<double> overall_best_clustering_score = clustering_score();
	printf("--- %u clusters --- ", (unsigned) overall_best_clustering.size());
	printf("score: %.2f = %.2f + %.2f --- %is ---\n", overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third, running_time);
	map<int, Attrset >::iterator misiit;
	for (misiit=overall_best_clustering.begin(); misiit!=overall_best_clustering.end(); ++misiit) {
		Attrset & cluster = misiit->second;
		for (int k=0; k<cluster.size; ++k)
			printf("%i ", cluster[k]);
		printf("\n");
	}
	printf("--- %u clusters --- ", (unsigned) overall_best_clustering.size());
	printf("score: %.2f = %.2f + %.2f --- %is ---\n", overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third, running_time);
	
	return;
}



/* Store the (current) 'overall_best_clustering' in 'outfilename'
   returns true on success
 */
bool AttrClusterer::save_clustering(string outfilename)
{
	string extension;
	size_t pointpos = outfilename.find_last_of('.');
	if (pointpos == string::npos)
		extension = "";
	else
		extension = outfilename.substr(pointpos+1, outfilename.size()-pointpos-1);
	
	if (extension == "xml")
		resultformat = FMT_XML;
	else
		resultformat = FMT_PLAINTXT;
	
	
	if (resultformat == FMT_PLAINTXT)
	{
		FILE *outfile = fopen(outfilename.c_str(), "w");
		
		fprintf(outfile, "#Attribute clustering of '%s'\n", datafilename.c_str());
		fprintf(outfile, "# attributes   : %d\n" , attr_count);
		fprintf(outfile, "# transactions : %d\n" , trans_count);
		fprintf(outfile, "# canon desclen: %.2f\n", canonical_descrlen());
		fprintf(outfile, "# clusters     : %d\n" , int(overall_best_clustering.size()));
		fprintf(outfile, "# score type   : %s\n", scoretypenames[scoretype].c_str());
		fprintf(outfile, "# search strat : %s\n", searchstrategynames[searchstrategy].c_str());
		fprintf(outfile, "# parameters   : ");
		fprintf(outfile, "k: %d\t",          param_k);
		fprintf(outfile, "stop: %d\t",       param_stop);
		fprintf(outfile, "beamwidth: %d\t",  param_beam_width);
		fprintf(outfile, "lambda: %f\t",     param_lambda);
		fprintf(outfile, "iterations: %d\t", param_iterations);
		fprintf(outfile, "cycles: %d",       param_cycles);
		fprintf(outfile, "\n");
		fprintf(outfile, "# score        : L(A,D)=%.2f\tL(A)=%.2f\tL(D|A)=%.2f\n", overall_best_clustering_score.first, overall_best_clustering_score.second, overall_best_clustering_score.third);
		fprintf(outfile, "# time         : %i\n", running_time);
		
		map<int, Attrset >::iterator misiit;
		for (misiit=overall_best_clustering.begin(); misiit!=overall_best_clustering.end(); ++misiit) {
			Attrset & cluster = misiit->second;
			for (int k=0; k<cluster.size; ++k)
				fprintf(outfile, "%d ", cluster[k]);
			fprintf(outfile, "\n");
		}
		fclose(outfile);
		
		return true;
	}
	else if (resultformat == FMT_XML)
	{
		xml_document doc;
		xml_node resultsnode; // node to store the result clustering under
		
		int outfilestatus = 0;
		string newoutfilename = outfilename;
		int cntr = 0;
		
		while (outfilestatus == 0)
		{
			if (cntr==0)
				newoutfilename = outfilename;
			else
				newoutfilename = outfilename + int_2_string(cntr);
			
			xml_parse_result res = doc.load_file(newoutfilename.c_str());
			
			if (res.status == status_file_not_found)
			{
				outfilestatus = 1;
				xml_node firstnode, datanode, attributesnode, attrnode;
				
				const char root [] = "<attributeclustering></attributeclustering>";
				doc.load_buffer(root, sizeof(root));
				
				firstnode = doc.root().child("attributeclustering");
				datanode = firstnode.append_child("data");
				datanode.append_child("filename").append_child(node_pcdata).set_value(datafilename.c_str());
				attributesnode = datanode.append_child("attributes");
				attributesnode.append_attribute("count") = attr_count;
				for (map<int, set<int> >::iterator mit=attr_items.begin(); mit!=attr_items.end(); ++mit) {
					attrnode = attributesnode.append_child("attribute");
					attrnode.append_attribute("number") = int_2_string(mit->first).c_str();
					attrnode.append_attribute("domainsize") = int_2_string(mit->second.size()).c_str();
				}
				datanode.append_child("transactions").append_child(node_pcdata).set_value(int_2_string(trans_count).c_str());
				datanode.append_child("canonDescrLen").append_child(node_pcdata).set_value(double_2_string(canonical_descrlen()).c_str());
				
				resultsnode = firstnode.append_child("results");
				
			}
			else if (res.status == status_ok) {
				if (datafilename == doc.root().child("attributeclustering").child("data").child("filename").child_value()) {
					outfilestatus = 2;
					resultsnode = doc.root().child("attributeclustering").child("results");
				}
			}
			else { // some error
				return false;
			}
			++cntr;
		}
		
		if (outfilename != newoutfilename) {
			outfilename = newoutfilename;
			if (verbosity >= 1)
				cerr << "WARNING: saved output to different filename: " << newoutfilename << endl;
		}
		
		// insert the (new) results
		xml_node insertnode = resultsnode.append_child("clustering");
		insertnode.append_attribute("method")       = searchstrategynames[searchstrategy].c_str();
		insertnode.append_attribute("scoretype")    = scoretypenames[scoretype].c_str();
		insertnode.append_attribute("score")        = double_2_string(overall_best_clustering_score.first).c_str();
		insertnode.append_attribute("modelscore")   = double_2_string(overall_best_clustering_score.second).c_str();
		insertnode.append_attribute("datascore")    = double_2_string(overall_best_clustering_score.third).c_str();
		insertnode.append_attribute("nrOfClusters") = int_2_string(int(overall_best_clustering.size())).c_str();
		insertnode.append_attribute("time")         = int_2_string(running_time).c_str();
		
		switch(searchstrategy)
		{
			case SRCH_HIERARCHICAL:
				insertnode.append_attribute("param_k")  = int_2_string(param_k).c_str();
				insertnode.append_attribute("param_stop")  = int_2_string(param_stop).c_str();
				break;
			case SRCH_BEAM:
				insertnode.append_attribute("param_k")  = int_2_string(param_k).c_str();
				insertnode.append_attribute("param_stop")  = int_2_string(param_stop).c_str();
				insertnode.append_attribute("param_beam_width")  = int_2_string(param_beam_width).c_str();
				break;
			case SRCH_ANNEAL:
				insertnode.append_attribute("param_lambda")  = double_2_string(param_lambda).c_str();
				insertnode.append_attribute("param_iterations")  = int_2_string(param_iterations).c_str();
				insertnode.append_attribute("param_cycles")  = int_2_string(param_cycles).c_str();
				break;
			case SRCH_INPUT:
				break;
			default:
				break;
		}
		
		map<int, Attrset >::iterator misiit;
		for (misiit=overall_best_clustering.begin(); misiit!=overall_best_clustering.end(); ++misiit) {
			Attrset & cluster = misiit->second;
			string list = "";
			for (int k=0; k<cluster.size; ++k)
				list += int_2_string(cluster[k]) + " ";
			insertnode.append_child("cluster").append_child(node_pcdata).set_value(list.c_str());
		}
		
		doc.save_file(outfilename.c_str());
		
		return true;
	}
	
	if (verbosity >= 1)
		printf("Saved result to '%s'\n", outfilename.c_str());
	
	return false;
}

