#include "similaritymatrix.h"
#include <algorithm>
#include <limits>
#include <iostream>


SimilarityMatrix::SimilarityMatrix(int n) : n(n)
{
	sim_heap = new SimHeapNode[(n*n + n) / 2];
	heap_size = 0;
	
	sim_matrix = new int*[n];
	for (int i=0; i<n; ++i)
		sim_matrix[i] = new int[n];
	
	for (int i=0; i<n; ++i)
		init_row(i);
	
	return;
}


SimilarityMatrix::SimilarityMatrix(const SimilarityMatrix & other) : n(other.n)
{
	sim_heap = new SimHeapNode[(n*n+n) / 2];
	std::copy(other.sim_heap, other.sim_heap+other.heap_size, sim_heap);
	heap_size = other.heap_size;
	
	sim_matrix = new int*[n];
	for (int i=0; i<n; ++i)
		sim_matrix[i] = new int[n];
	for (int i=0; i<n; ++i)
		init_row(i);
	
	for (int i=0; i<n; ++i)
		for (int j=0; j<n; ++j)
			sim_matrix[i][j] = other.sim_matrix[i][j];
	
	indices = set<int>(other.indices);
	
	return;
}


SimilarityMatrix::~SimilarityMatrix()
{
	delete [] sim_heap;
	
	for (int i=0; i<n; ++i)
		delete [] sim_matrix[i];
	delete [] sim_matrix;
	
	return;
}


void SimilarityMatrix::init_row(int i)
{
	indices.insert(i);
	
	return;
}


void SimilarityMatrix::insert(int i, int j, double sim)
{
	if (i==j) return;
	if (i > j) std::swap(i,j);
	
	int pos = heap_size;
	int parentpos = (pos-1) >> 1;
	while (pos && sim > sim_heap[parentpos].sim) {
		move_node(parentpos, pos);
		pos = parentpos;
		parentpos = (pos-1) >> 1;
	}
	
	sim_heap[pos].sim = sim;
	sim_heap[pos].i = i;
	sim_heap[pos].j = j;
	
	sim_matrix[i][j] = pos;
	sim_matrix[j][i] = pos;
	
	++heap_size;
	
	return;
}


void SimilarityMatrix::delete_row(int i)
{
	set<int>::iterator ind;
	for (ind = indices.begin(); ind != indices.end(); ++ind)
		remove(i, *ind);
	
	indices.erase(i);
	
	return;
}


void SimilarityMatrix::remove(int i, int j)
{
	if (i==j) return;
	if (i > j) std::swap(i, j);
	
	int pos = sim_matrix[i][j];
	int parentpos = (pos-1) >> 1;
	
	while (pos) {
		move_node(parentpos, pos);
		pos = parentpos;
		parentpos = (pos-1) >> 1;
	}
	
	move_node(heap_size-1, 0);
	--heap_size;
	
	SimHeapNode s = sim_heap[0];
	
	int kid1pos = 1;
	int kid2pos = 2;
	while (((kid1pos < heap_size) && (s.sim < sim_heap[kid1pos].sim)) || ((kid2pos < heap_size) && (s.sim < sim_heap[kid2pos].sim))) {
		if ( kid2pos == heap_size || sim_heap[kid1pos].sim > sim_heap[kid2pos].sim ) {
			move_node(kid1pos, pos);
			pos = kid1pos;
		}
		else {
			move_node(kid2pos, pos);
			pos = kid2pos;
		}
		kid1pos = (pos << 1) + 1;
		kid2pos = kid1pos + 1;
	}
	
	sim_heap[pos] = s;
	sim_matrix[s.i][s.j] = pos;
	sim_matrix[s.j][s.i] = pos;
	
	return;
}


pair<double, pair<int,int> > SimilarityMatrix::get_max() const
{
	return pair<double, pair<int,int> >(sim_heap[0].sim, pair<int,int>(sim_heap[0].i, sim_heap[0].j));
}


pair<double, pair<int,int> > * SimilarityMatrix::get_topk(int k) const
{
	pair<double, pair<int,int> > * topk = new pair<double, pair<int,int> >[k];
	int len = 0;
	
	for (int i=len; i<k; ++i)
		topk[i].first = std::numeric_limits<double>::quiet_NaN();
	
	get_topk(k, 0, topk, len);
	
	return topk;
}


void SimilarityMatrix::get_topk(int k, int pos, pair<double, pair<int,int> > * topk, int & len) const
{
	if (pos >= heap_size)
		return;
	
	if (len < k) {
		topk[len] = pair<double, pair<int,int> >(sim_heap[pos].sim, pair<int,int>(sim_heap[pos].i, sim_heap[pos].j));
		++len;
		std::sort(topk, topk+len);
		std::reverse(topk, topk+len);
		get_topk(k, (pos << 1) + 1, topk, len);
		get_topk(k, (pos << 1) + 2, topk, len);
	}
	else if (sim_heap[pos].sim > topk[k-1].first) {
		topk[k-1] = pair<double, pair<int,int> >(sim_heap[pos].sim, pair<int,int>(sim_heap[pos].i, sim_heap[pos].j));
		std::sort(topk, topk+len);
		std::reverse(topk, topk+len);
		get_topk(k, (pos << 1) + 1, topk, len);
		get_topk(k, (pos << 1) + 2, topk, len);
	}
	
	return ;
}

