/*
 * Template Trie structure
 * Author: Michael Mampaey
 */

#include "intmap.h"
#include <algorithm>
#include <math.h>
#include <iostream>
#include "auxiliary.h"
#include <utility>

using namespace std;



template <class T> class Trie;


template <class T> class Node
{
public:
	Node(T val);
	~Node();
	
	typedef T value_type;
	
private:
	T val;
	IntMap<Node<T>* > children;
	
	friend class Trie<T>;
};



template <class T> class Trie
{
public:
	Trie(T defaultval = T(0));
	~Trie();
	
	Node<T> *increment(const int *key, int keylen);
	mypair<double> get_entropy_and_logsum(Node<T>*start=0) const;
	double map_sum(double (*function)(T), Node<T>*start=0) const;
	double get_refined_entropy(double domsize, Node<T>*start=0) const;
	int get_size() const { return size; };
	
	typedef T value_type;
	
private:
	Node<T> *root;
	double sum;
	int size;
	
	const T defaultval;
	
};




template <class T> inline Node<T>::Node(T val): val(val)
{

}


template <class T> inline Node<T>::~Node()
{
	for (int i=0; i<children.size; ++i) {
		delete children.map[i].second;
	}
}



template <class T> Trie<T>::Trie(T defaultval) : defaultval(defaultval)
{
	root = new Node<T>(defaultval);
	size = 0;
	sum = 0.0;
}


template <class T> Trie<T>::~Trie()
{
	delete root;
}


template <class T> Node<T>* Trie<T>::increment(const int *key, int keylen)
{
	Node<T> *current = root;
	
	Node<T>* pit;
	int *sk = (int*) key;
	while (sk!=key+keylen) {
		pit = current->children.get(*sk);
		if (pit==0) {
			Node<T> *newnode = new Node<T>(defaultval);
			current->children.insert(*sk, newnode);
			current = newnode;
		}
		else
			current = pit;
		++sk;
	}
	
	if (current->val==0)
		++size;
	current->val += 1;
	++sum;
	
	return current;
}


template <class T> inline mypair<double> Trie<T>::get_entropy_and_logsum(Node<T> *start) const
{
	static double f;
	Node<T> *current = start;
	if (current==0)
		current = root;
	
	mypair<double> hls = mypair<double>(0.0, 0.0);
	
	if (current->val > 0) {
		f = log_2(current->val / sum);
		hls.first -= current->val / sum * f;
		hls.second -= f;
	}
	
	for (int i=current->children.size-1; i>=0; --i)
		hls += get_entropy_and_logsum(current->children.map[i].second);
	
	return hls;
}



/* Return the sum of 'function' applied to the elements of the tree
 */
template <class T> inline double Trie<T>::map_sum(double (*function)(T), Node<T> *start) const
{
	Node<T> *current = start;
	if (current==0)
		current = root;
	
	double mapsum = function(current->val);
	for (int i=current->children.size-1; i>=0; --i)
		mapsum += map_sum(function, current->children.map[i].second);
	
	return mapsum;
}



template <class T> inline double Trie<T>::get_refined_entropy(double domsize, Node<T> *start) const
{
	double re = 0.0;
	
	Node<T> *current = start;
	if (current==0) {
		current = root;
		if (0.5*domsize == floor(0.5*domsize))
			re += log_factorial(sum + 0.5*domsize - 1) - log_factorial(0.5*domsize - 1);
		else
			re += log_factorial(2*sum + domsize-1.0) - log_factorial(sum + floor(0.5*domsize)) - 2*sum - log_factorial(domsize-1.0) + log_factorial(floor(0.5*domsize));
	}
	
	if (current->val > 0)
		re -= log_factorial(2 * current->val) - log_factorial(current->val) - 2 * current->val;
	for (int i=current->children.size-1; i>=0; --i)
		re += get_refined_entropy(domsize, current->children.map[i].second);
	
	return re;
}

